@@ -5,6 +5,9 @@ from matplotlib.backend_bases import key_press_handler
 from matplotlib.cbook import CallbackRegistry
 from matplotlib.figure import Figure
 
+from ._util import ConflictingArgumentsError, list_to_onehot, onehot_to_list
+from ._widgets import button_array
+
 
 def gen_key_press_handler(skip_keys):
     def handler(event, canvas=None, toolbar=None):
@@ -20,8 +23,11 @@ class image_labeller:
         self,
         images,
         classes,
-        init_labels=None,
+        multiclass=False,
         label_keymap: Union[List[str], str] = "1234",
+        title=None,
+        init_labels=None,
+        init_labels_onehot=None,
         labelling_advances_image: bool = True,
         N_images=None,
         fig: Figure = None,
@@ -33,17 +39,30 @@ class image_labeller:
         images : (N, Y, X) ArrayLike
         classes : (N,) ArrayLike
             The available classes for the images.
-        init_labels: 1D ArrayLike, optional
-            The initial labels for the images. If given it must be the same length as
-            *images*
+        multiclass : bool, default: False
+            Whether to allow for an image to have multiple classes or just one.
         label_keymap : list of str, or str
             If a str must be one of the predefined values *1234* (1, 2, 3,..),
             *qwerty* (q, w, e, r, t, y). If an iterable then the items will be assigned
             in order to the classes. WARNING: These keys will be removed from the
             default keymap for that figure. So if *s* is included then *s* will no
             longer perform savefig.
+        title : str or Callable, optional
+            A {} style format string for the title of images or a function that
+            returns a title string given the image index. If a format string
+            then it will be used with ``title.format(image_index=...)``. If *None*
+            then the default str of 'Image Index: {image_index}' will be used.
+        init_labels: list of list or list of str, optional
+            The initial labels for the images. If given it must be the same length as
+            *images* and each entry should be either a single class or an iterable of
+            classes. Incompatible with *init_labels_onehot*.
+        init_labels_onehot: 2D ArrayLike, optional
+            The initial labels for the images as a onehot encoding. If given it must
+            have shape (N_images, N_classes) and be castable to a boolean array.
+            Incompatible with *init_labels*.
         labelling_advances_image : bool, default: True
             Whether labelling an image should advance to the next image.
+            Ignored if *multiclass* is True.
         N_images : int or None
             The number of images. Required if passing a Callable for images, otherwise
             ignored.
@@ -53,7 +72,21 @@ class image_labeller:
         **imshow_kwargs :
             kwargs to be passed to the imshow function that displays the images.
         """
+        self._multi = multiclass
         self._images = images
+        if title is None:
+            title = "Image Index: {image_index}"
+        if isinstance(title, str):
+
+            def _title(image_index):
+                return title.format(image_index=image_index)
+
+            self._title = _title
+        elif callable(title):
+            self._title = title
+        else:
+            raise TypeError("Title must be a str or a Callable")
+
         if callable(images):
             if not isinstance(N_images, int):
                 raise TypeError(
@@ -74,12 +107,9 @@ class image_labeller:
 
         self._label_advances = labelling_advances_image
 
-        if init_labels is None:
-            self._labels = [None] * self._N_images
-        elif len(init_labels) != self._N_images:
-            raise ValueError("init_labels must have the same length as images")
-        else:
-            self._labels = init_labels
+        # TODO: sync this up with labels
+        # TODO: make sure init_labels does something here
+        self._onehot = np.zeros((self._N_images, len(classes)), dtype=bool)
 
         if label_keymap == "1234":
             if len(classes) > 10:
@@ -95,10 +125,29 @@ class image_labeller:
                     "please provide a custom keymap"
                 )
             self._label_keymap = {"qwertyuiop"[c]: c for c in range(len(classes))}
+        elif len(label_keymap) != len(classes):
+            raise ValueError("label_keymap must have the same length as classes")
         else:
             self._label_keymap = {label_keymap[i]: i for i in range(len(label_keymap))}
 
-        self._classes = classes
+        # make array for easy indexing
+        self._classes = np.asarray(classes)
+
+        if init_labels is not None and init_labels_onehot is not None:
+            raise ConflictingArgumentsError(
+                "init_labels and init_labels_onehot cannot both be *None*"
+            )
+
+        if init_labels is not None:
+            # length errors are handled in the setter
+            self.labels = init_labels
+        elif init_labels_onehot is not None:
+            self.labels_onehot = init_labels_onehot
+        else:
+            if self._multi:
+                self.labels = [[]] * self._N_images
+            else:
+                self.labels = [None] * self._N_images
 
         if fig is None:
             import matplotlib.pyplot as plt
@@ -116,59 +165,102 @@ class image_labeller:
         )
 
         self._image_index = 0
-        self._ax = self._fig.add_subplot(111)
+        if self._multi:
+            self._image_ax, self._info_ax = self._fig.subplots(1, 2)
+        else:
+            gs = self._fig.add_gridspec(1, 5)
+            self._image_ax = self._fig.add_subplot(gs[:, :-1])
+            self._info_ax = self._fig.add_subplot(gs[:, -1])
+
+        self._info_ax.axis("off")
         aspect = imshow_kwargs.pop("aspect", "equal")
-        self._im = self._ax.imshow(self._get_image(0), aspect=aspect, **imshow_kwargs)
+        self._im = self._image_ax.imshow(
+            self._get_image(0), aspect=aspect, **imshow_kwargs
+        )
+
+        if self._multi:
+
+            def on_state_change(new_state, old_state):
+                self._onehot[self._image_index] = new_state
+                # self.labels[self._image_index] = self._classes[new_state]
+
+            texts = []
+            for key, klass in zip(self._label_keymap.keys(), classes):
+                texts.append(f"[{key}]\n{str(klass)}")
+            self._buttons = button_array(texts, self._info_ax)
+            self._buttons.on_state_change(on_state_change)
+        else:
+            # shift axis to make room for list of keybindings
+            # box = self._image_ax.get_position()
+            # box.x0 = box.x0 - 0.20
+            # box.x1 = box.x1 - 0.20
+            # self._image_ax.set_position(box)
+
+            # these are matplotlib.patch.Patch properties
+            props = dict(boxstyle="round", facecolor="wheat", alpha=0.5)
+
+            textstr = """Keybindings
+            <- : Previous Image
+            -> : Next Image"""
+            horiz_pos = 0.575
+
+            self._info_ax.text(
+                horiz_pos,
+                0.75,
+                textstr,
+                # transform=self._fig.transFigure,
+                fontsize=14,
+                verticalalignment="top",
+                bbox=props,
+                horizontalalignment="left",
+            )
+
+            textstr = """Class Keybindings:\n"""
+            for k, v in self._label_keymap.items():
+                textstr += f"{k} : {self._classes[v]}\n"
+
+            self._info_ax.text(
+                horiz_pos,
+                0.55,
+                textstr,
+                # transform=self._fig.transFigure,
+                fontsize=14,
+                verticalalignment="top",
+                bbox=props,
+            )
+
+            textstr = f"Current Class:\n{str(self._labels[0])}"
+            self._class_display = self._info_ax.text(
+                horiz_pos,
+                0.25,
+                textstr,
+                # transform=self._fig.transFigure,
+                fontsize=14,
+                verticalalignment="top",
+                bbox=props,
+            )
 
-        # shift axis to make room for list of keybindings
-        box = self._ax.get_position()
-        box.x0 = box.x0 - 0.20
-        box.x1 = box.x1 - 0.20
-        self._ax.set_position(box)
         self._update_title()
 
-        # these are matplotlib.patch.Patch properties
-        props = dict(boxstyle="round", facecolor="wheat", alpha=0.5)
-
-        textstr = """Keybindings
-        <- : Previous Image
-        -> : Next Image"""
-
-        self._ax.text(
-            1.05,
-            0.95,
-            textstr,
-            transform=self._ax.transAxes,
-            fontsize=14,
-            verticalalignment="top",
-            bbox=props,
-            horizontalalignment="left",
-        )
-
-        textstr = """Class Keybindings:\n"""
-        for k, v in self._label_keymap.items():
-            textstr += f"{k} : {self._classes[v]}\n"
-
-        self._ax.text(
-            1.05,
-            0.55,
-            textstr,
-            transform=self._ax.transAxes,
-            fontsize=14,
-            verticalalignment="top",
-            bbox=props,
-        )
-
         self._fig.canvas.mpl_connect("key_press_event", self._key_press)
         self._observers = CallbackRegistry()
 
     @property
     def ax(self):
-        return self._ax
+        """
+        **readonly** - The `~matplotlib.axes.Axes` object the image's are displayed on.
+        """
+        return self._image_ax
 
     @property
     def labels(self):
-        return self._labels
+        """
+        The current labels as a list of lists or a list of strings.
+        """
+        if self._multi:
+            return onehot_to_list(self._onehot, self._classes)
+        else:
+            return self._labels
 
     @labels.setter
     def labels(self, value):
@@ -176,10 +268,40 @@ class image_labeller:
             raise ValueError(
                 "Length of labels must be the same as the number of images"
             )
-        self._labels = value
+        if self._multi:
+            self._onehot = list_to_onehot(value, self._classes)
+        else:
+            self._labels = value
+
+    @property
+    def labels_onehot(self):
+        """
+        The current labels as a one hot encoding.
+        """
+        if self._multi:
+            return self._onehot
+        else:
+            return list_to_onehot(self._labels, self._classes)
+
+    @labels_onehot.setter
+    def labels_onehot(self, value):
+        value = np.asanyarray(value)
+        expected_shape = (self._N_images, len(self._classes))
+        if value.shape != expected_shape:
+            raise ValueError(
+                "One hot labels must have shape (N images, num classes."
+                f"Expected shape {expected_shape} but got {value.shape}"
+            )
+        if self._multi:
+            self._onehot = value
+        else:
+            self._labels = onehot_to_list(value)
 
     @property
     def image_index(self):
+        """
+        **int** the index of the currently displayed image.
+        """
         return self._image_index
 
     @image_index.setter
@@ -202,18 +324,25 @@ class image_labeller:
         self._update_displayed()
 
     def _update_title(self):
-        self._ax.set_title(
-            f"Image {self._image_index}\nLabel: {self._labels[self._image_index]}"
-        )
+        text = self._title(self._image_index)
+        self._image_ax.set_title(text)
 
     def _update_displayed(self):
         image = np.asarray(self._get_image(self._image_index))
         # for some reason this keeps getting turned off by something
-        self._ax.set_autoscale_on(True)
+        self._image_ax.set_autoscale_on(True)
         self._im.set_data(image)
         self._im.set_extent((-0.5, image.shape[1] - 0.5, image.shape[0] - 0.5, -0.5))
         self._update_title()
         self._observers.process("image-changed", self._image_index, image)
+        if self._multi:
+            with self._buttons.no_callbacks():
+                # TODO: check that this no_callbacks actually works....
+                new_state = self._onehot[self._image_index]
+                self._buttons.set_states(new_state)
+        else:
+            textstr = f"Current Class:\n{str(self._labels[self._image_index])}"
+            self._class_display.set_text(textstr)
         self._fig.canvas.draw_idle()
 
     def _key_press(self, event):
@@ -222,13 +351,21 @@ class image_labeller:
         elif event.key == "right":
             self.image_index += 1
         elif event.key in self._label_keymap:
-            klass = self._classes[self._label_keymap[event.key]]
-            self._labels[self._image_index] = klass
+            which_label = self._label_keymap[event.key]
+            klass = self._classes[which_label]
+            if self._multi:
+                img_labels = self._onehot[self._image_index]
+                img_labels[which_label] = not img_labels[which_label]
+                self._buttons.set_states(img_labels)
+            else:
+                self._labels[self._image_index] = klass
             self._observers.process("label-assigned", self._image_index, klass)
-            if self._label_advances:
+            if self._label_advances and not self._multi:
                 if self.image_index == self._N_images - 1:
                     # make sure we update the title we are on the last image
                     self._update_title()
+                    textstr = f"Current Class:\n{str(self._labels[self._image_index])}"
+                    self._class_display.set_text(textstr)
                     self._fig.canvas.draw_idle()
                 else:
                     self.image_index += 1
